/*
//    FOBO autonomous navigation program. Version 2.0
//      This program makes FOBO walk forward until it observes an obstacle within 
//      5 inces at which point it will turn right until the way is clear and then
//      start  walking forward again (repeat).
//
//    Copyright (C) 2012  Jonathan Dowdall, Project Biped (www.projectbiped.com)
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// this is the library for the servo shield board ... it can be found at http://www.renbotics.com/files/servoshield.zip
// for you program to compile you need to unzip the ServoShield folder and put the whole folder under you "libraries"
// folder inside of your arudino development environment folder ... example "C:\Program Files\arduino-0022\libraries\ServoShield"
#include <ServoShield.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Constants
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const int pingPin                    = 10;     //digitial pin number on the arduino board that has the ping data line plugged into it
const int maximumServoShieldPosition = 2000;   // the maximum pulse duration for the servo shield (2ms pulse)
const int minimumServoShieldPosition = 1000;   // the minimum pulse duration for the servo shield (1ms pulse)
const int numberOfServos             = 8;      // the communication with the control program is expecting 8 servos 

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class Action
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  The Action class collects information necessary to perform a sequence of movements.  It allows for intermediate 
  //  frames (robot positions) to be generated between key frames on the fly (keeping memory usage low).   
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////  
  
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //Members
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  public:
    int numberOfFrames;      //total number of frames in the action
    int numberOfKeyFrames;   //number of key frames (this must match the first dimension of the frames array)
    int frameDelay;          //number of milliseconds to wait inbetween each frame during playback
    int frameNumber;         //current frame number during playback
    int* frames;             //pointer to a two dimensional array containing the individual frames

  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //Methods
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  Action(int setNumberOfKeyFrames, int setNumberOfFrames, int setFrameDelay, void* setFrames)
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  {
    frames            = (int*)setFrames;
    numberOfKeyFrames = setNumberOfKeyFrames;
    numberOfFrames    = setNumberOfFrames;
    frameDelay        = setFrameDelay;
    frameNumber       = 0;
  }  
  
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void NextFrame()
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  {
    frameNumber++;
    if(frameNumber >= numberOfFrames)
      frameNumber = 0;
      
    delay(frameDelay);    
  }

  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void GetCurrentFrame(int* frame)
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  {
     GetFrame(frame, frameNumber);
  }
   
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void GetFrame(int* frame, int targetFrameNumber)
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  {
    //Compute the frame based on the target frame number.  Remember that this can either be a key frame or an intermediate
    //frame between two key frames.
    
    //compute the number of intermediate frames between each key frame
    int numberOfIntermediateFrames = numberOfFrames/numberOfKeyFrames;
    
    //compute the closest key frame before the target frame number
    int sourceKeyFrame = targetFrameNumber/numberOfIntermediateFrames;
    
    //get the key frame after the source key frame 
    int destinationKeyFrame = sourceKeyFrame + 1;
    //wrap around if this is the last frame
    if(destinationKeyFrame >= numberOfKeyFrames)
      destinationKeyFrame = 0;
      
    //compute mixing percentage between the source and destination key frame
    float percent = (float)(targetFrameNumber - sourceKeyFrame*numberOfIntermediateFrames)/(float)numberOfIntermediateFrames;
    
    //mix the source and destination key frame to produce the target frame
    GetIntermediateFrame(frames + sourceKeyFrame*numberOfServos, frames + destinationKeyFrame*numberOfServos, percent, frame);    
  }
  
  private:
  
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void GetIntermediateFrame(int* fromState, int* toState, float percent, int* outState)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  {
      //update each servo
      for (int servo = 0; servo < numberOfServos; servo++)
      {   
        //compute the angles from both the from state and the to state for this servo.
        //remember that the states are stored as 100*angle ... so they need to be converted back into angles
        int source = ((float)fromState[servo]);      
        int target = ((float)toState[servo]);    
        
        //the servo angle used is the linear interpolation between the two 
        outState[servo] = percent*(target - source) + source;
      }
  }
  
};
 
 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Action Frames
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//Joints positions are in degrees * 100 (home position is 60 degrees)
/////////////////////////////////////////////
//JOINT INDEXES
// 0 right ankle
// 1 right lower leg
// 2 right upper leg
// 3 right hip
// 4 left ankle
// 5 left lower leg
// 6 left upper leg
// 7 left hip
/////////////////////////////////////////////
int walkFrames[10][8] =
{
                      {6900, 8100, 8700, 6450, 9599, 8400, 7500, 8700 },
                      {3000, 7500, 9750, 3299, 6000, 8100, 6600, 6000 },
                      {1200, 5400, 8400, 1800, 4500, 8100, 5700, 4500 },
                      {900, 4800, 6600, 3600, 4200, 7200, 6300, 4500 },
                      {1800, 5100, 5100, 5700, 4800, 5100, 5400, 6300 },
                      {2500, 3600, 4500, 3300, 5100, 3900, 3300, 5600 },
                      {6000, 3900, 5400, 6000, 9000, 4500, 2300, 8800 },
                      {7500, 3900, 6300, 7500, 10800, 6600, 3600, 10200 },
                      {7800, 4800, 5700, 7500, 11100, 7200, 5400, 8400 },
                      {7200, 6900, 6600, 5700, 10200, 6900, 6900, 6300 }
  
};

int turnFrames[7][8] = {
                      { 6000,  6000,  6000,  6000,  6000,  6000,  6000,  6000 },
                      { 5550,  5700,  7500,  5700,  6300,  6300,  5100,  6600 },
                      { 3299,  5700,  7500,  4500,  6000,  6300,  5100,  6600 },
                      { 1500,  5700,  7500,  3600,  5700,  6600,  3900,  5400 },
                      {  600,  5700,  7500,  3600,  4500,  6900,  4200,  5100 },
                      {  600,  5700,  6300,   600,  2700,  5400,  1800,  3000 },
                      { 6000,  6000,  6000,  6000,  6000,  6000,  6000,  6000 }
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Calibration
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const int calibration[8] = {-573, 723, 823, 149, 474, -374, -1621, -498 };  //contains the offset for each servo to make it align with the default position

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Variables
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ServoShield servos;                  //servoShield object to interact with the servo board
int         distance         = 100;  //distance measured on the ping
int         lastDistance     = 100;  //previous distance measured on the ping

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Actions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Action turn(7, 35, 50, turnFrames);   //the turn right action
Action walk(10, 100, 25, walkFrames); //the walk forward action
Action* action;                       //pointer to the current action

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void setup()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //set the initial action
  action = &walk;
  
  //wait for a second to begin (keeps the communication line open in case a new program is being downloaded)
  delay(1000);    
  
  //start up the communication
  Serial.begin(9600);  
  
  //initialize the servos
  initializeServos();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void initializeServos()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{   
  //Initialize the servos (FOBO only uses 8) but the control board could do up to 16
  //Set the minimum and maximum pulse duration  
  for (int servo = 0; servo < numberOfServos; servo++)
    servos.setbounds(servo, minimumServoShieldPosition, maximumServoShieldPosition); 
    
  //Start the servo shield
  servos.start(); 
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ObstacleInPath()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //initialize the result
  boolean foundObstacle = true;
 
  //measure the distance 
  distance = MeasureDistance();
  
  //ignore error readings 
  if(distance == 0)
    distance = lastDistance;

  //make sure the object was observed at least twice in a row (high frequency filter)
  if(distance > 5 && lastDistance > 5)    
    foundObstacle = false;
        
  //remember the distance 
  lastDistance = distance;
    
  //return the result
  return foundObstacle;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void TransitionTo(int* fromState, int* toState, int transitionFrames, int wait)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //the idea is to start at the fromState and move to the toState with some intermediate states (transitionFrames) with a delay at each transition state
  int frame[numberOfServos];
  
  //walk through each transition frame (skipping the first one because this is the fromState which the robot should already be at  
  for(int f = 1; f < transitionFrames; f++)
  {
    //set the servo positions
    //SetIntermediateFrame(fromState, toState, (float)f/(float)transitionFrames);
    GetIntermediateFrame(fromState, toState, (float)f/(float)transitionFrames, frame);
    SetServoPositions(frame);
    
    //wait for the specified number of milliseconds
    delay(wait);
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void GetIntermediateFrame(const int* fromState, const int* toState, float percent, int* outState)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
    //update each servo
    for (int servo = 0; servo < numberOfServos; servo++)
    {   
      //compute the angles from both the from state and the to state for this servo.
      //remember that the states are stored as 100*angle ... so they need to be converted back into angles
      int source = ((float)fromState[servo]);      
      int target = ((float)toState[servo]);    
      
      //the servo angle used is the linear interpolation between the two 
      outState[servo] = percent*(target - source) + source;
    }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void UpdateAction()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //check to see if the robot is walking
  if(action == &walk)
  {
    //look for obstacles in the way ... but only respond if the robot is in a frame (either left or right) where it can transition to turn
    if( ObstacleInPath() && (action->frameNumber == 0 || action->frameNumber == 49))
    {
      //transition to the start of the turning action
      int sourceFrame[8];
      walk.GetCurrentFrame(sourceFrame);
      int destinationFrame[8];
      turn.GetFrame(destinationFrame, 0);
      TransitionTo(sourceFrame, destinationFrame, 10, 40);
      
      //next time the robot starts walking it will start with its weight on the right foot.
      //starting at frame 0 makes the robot turn left a bit and it might run into the wall it was try to avoid
      action->frameNumber = 0;
      
      //make turning the current action
      action = &turn;
    }
    else
    {
      //no obstacle (or can't transition to a turn) so just move to the next frame in the walking sequence
      action->NextFrame();    
    }
  }
  //check to see if the robot is turning
  else if (action == &turn)
  {
    //update for turning action
    
    //check to see if this is the last frame in the turning action ... if so then look to see if the way is clear
    if(action->frameNumber == action->numberOfFrames -1 && !ObstacleInPath())
    {
      int sourceFrame[8];
      turn.GetFrame(sourceFrame, 0);
      int destinationFrame[8];
      walk.GetFrame(destinationFrame, 0);
      
      //no obstacle so start walking again
      TransitionTo(sourceFrame, destinationFrame, 10, 40);            
            
      //next time the turning action is started begin at frame 0
      action->frameNumber = 0;
      
      //change the action to walk
      action = &walk;      
    }
    else
    {
      //go to the next frame in the turning action
      action->NextFrame();
    }
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void loop()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //this is the main update loop for the microcontroller
  
  //update the current action
  UpdateAction();
  
  //get the frame from the current action  
  int frame[numberOfServos];
  action->GetCurrentFrame(frame);
  
  //set the servo positions for this frame number
  SetServoPositions(frame);                    
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void SetServoPositions(int* frame)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //update each servo
  for (int servo = 0; servo < numberOfServos; servo++)
  {   
    //the servo angle used is the linear interpolation between the two 
    float servoAngle = (frame[servo] + calibration[servo])/100.0f;
     
    // the servo control shield commands the servos via pulse width modulations, not angles
    // the PWMs range from 1000 (equal to 0 degrees) to 2000 (equal to 120 degrees)
    // so the servo angle needs to be converted to the corresponding PWM range.
    int pwm = (int)(servoAngle/120.0* (float)(maximumServoShieldPosition - minimumServoShieldPosition)) + minimumServoShieldPosition;
      
    //Set PWM for the servo shield to send out to the servo
    servos.setposition(servo, pwm);       
  }


}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// some of the code below was found on http://arduino.cc/en/Tutorial/Ping?from=Tutorial.UltrasoundSensor.  Thanks Arduino guys!
//  credit:
//    by David A. Mellis
//    modified 30 Aug 2011
//    by Tom Igoe
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
long MeasureDistance()
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // establish variables for duration of the ping, 
  // and the distance result in inches and centimeters:
  long duration, inches, cm;

  // The PING))) is triggered by a HIGH pulse of 2 or more microseconds.
  // Give a short LOW pulse beforehand to ensure a clean HIGH pulse:
  pinMode(pingPin, OUTPUT);
  digitalWrite(pingPin, LOW);
  delayMicroseconds(2);
  digitalWrite(pingPin, HIGH);
  delayMicroseconds(5);
  digitalWrite(pingPin, LOW);

  // The same pin is used to read the signal from the PING))): a HIGH
  // pulse whose duration is the time (in microseconds) from the sending
  // of the ping to the reception of its echo off of an object.
  pinMode(pingPin, INPUT);
  duration = pulseIn(pingPin, HIGH);

  // convert the time into a distance
  inches = microsecondsToInches(duration);
  cm = microsecondsToCentimeters(duration);

  return inches;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
long microsecondsToInches(long microseconds)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // According to Parallax's datasheet for the PING))), there are
  // 73.746 microseconds per inch (i.e. sound travels at 1130 feet per
  // second).  This gives the distance travelled by the ping, outbound
  // and return, so we divide by 2 to get the distance of the obstacle.
  // See: http://www.parallax.com/dl/docs/prod/acc/28015-PING-v1.3.pdf
  return microseconds / 74 / 2;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
long microsecondsToCentimeters(long microseconds)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // The speed of sound is 340 m/s or 29 microseconds per centimeter.
  // The ping travels out and back, so to find the distance of the
  // object we take half of the distance travelled.
  return microseconds / 29 / 2;
}

